/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: SSHTunnelManagerImpl.java 354 2011-09-29 09:57:24Z chazalet $
 *
 */

package org.ow2.sirocco.cloudmanager.sshtunnel.impl;

import java.util.Dictionary;
import java.util.HashMap;
import java.util.Map;

import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.ow2.sirocco.cloudmanager.sshtunnel.api.ISSHTunnelManager;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.UserInfo;

public class SSHTunnelManagerImpl implements ISSHTunnelManager, ManagedService {
    private static Log logger = LogFactory.getLog(SSHTunnelManagerImpl.class);

    private static final String PORT_RANGE_MIN_PROP_NAME = "service.sshtunnelmanager.portrange.min";

    private static final String PORT_RANGE_MAX_PROP_NAME = "service.sshtunnelmanager.portrange.max";

    private static final int DEFAULT_LOCAL_PORT_PORT_RANGE_MIN = 6000;

    private static final int DEFAULT_LOCAL_PORT_PORT_RANGE_MAX = 6999;

    private static class SSHTunnel {
        @SuppressWarnings("unused")
        String targetHost;

        @SuppressWarnings("unused")
        int targetPort;

        int localPort;

        Session session;

        SSHTunnel(final String targetHost, final int targetPort, final int localPort, final Session session) {
            this.targetHost = targetHost;
            this.targetPort = targetPort;
            this.localPort = localPort;
            this.session = session;
        }

    }

    private static class LocalUserInfo implements UserInfo {
        String passwd;

        LocalUserInfo(final String passwd) {
            this.passwd = passwd;
        }

        public String getPassword() {
            return this.passwd;
        }

        public boolean promptYesNo(final String str) {
            return true;
        }

        public String getPassphrase() {
            return null;
        }

        public boolean promptPassphrase(final String message) {
            return true;
        }

        public boolean promptPassword(final String message) {
            return true;
        }

        public void showMessage(final String message) {
        }
    }

    private JSch jsch;

    private int nextTunnelLocalPort;

    private Map<String, SSHTunnel> openTunnels = new HashMap<String, SSHTunnel>();

    private int portRangeMin = SSHTunnelManagerImpl.DEFAULT_LOCAL_PORT_PORT_RANGE_MIN;

    private int portRangeMax = SSHTunnelManagerImpl.DEFAULT_LOCAL_PORT_PORT_RANGE_MAX;

    SSHTunnelManagerImpl() {
    }

    public void start() {
        SSHTunnelManagerImpl.logger.info("SSH Tunnel Manager started");
        this.jsch = new JSch();
        this.nextTunnelLocalPort = this.portRangeMin;
    }

    @SuppressWarnings("rawtypes")
    public void updated(final Dictionary properties) throws ConfigurationException {
        if (properties != null) {
            try {
                String s = (String) properties.get(SSHTunnelManagerImpl.PORT_RANGE_MIN_PROP_NAME);
                if (s != null) {
                    this.portRangeMin = Integer.parseInt(s);
                }
                s = (String) properties.get(SSHTunnelManagerImpl.PORT_RANGE_MAX_PROP_NAME);
                if (s != null) {
                    this.portRangeMax = Integer.parseInt(s);
                }
            } catch (NumberFormatException ex) {
                SSHTunnelManagerImpl.logger.error("Illegal parameter port SSH tunnel manager port range " + ex.getMessage());
                this.portRangeMin = SSHTunnelManagerImpl.DEFAULT_LOCAL_PORT_PORT_RANGE_MIN;
                this.portRangeMax = SSHTunnelManagerImpl.DEFAULT_LOCAL_PORT_PORT_RANGE_MAX;
            }
        }
        this.nextTunnelLocalPort = this.portRangeMin;
        SSHTunnelManagerImpl.logger.info("SSH tunnel manager ready, port range [" + this.portRangeMin + "," + this.portRangeMax
            + "]");
    }

    public synchronized int createSSHTunnel(final String targetHost, final int targetPort, final String user,
        final String password, final String keyFile) throws Exception {
        if (keyFile != null) {
            this.jsch.removeAllIdentity();
            this.jsch.addIdentity(keyFile, "passphrase");
        }
        SSHTunnel tunnel = this.openTunnels.get(targetHost + targetPort);
        if (tunnel != null) {
            throw new Exception("Tunnel already created");
        }

        Session session = this.jsch.getSession(user, targetHost, 22);
        LocalUserInfo lui = new LocalUserInfo(password);
        session.setUserInfo(lui);
        session.connect();

        SSHTunnelManagerImpl.logger.debug("nextPort=" + this.nextTunnelLocalPort);
        int tunnelLocalPort = this.nextTunnelLocalPort;
        this.nextTunnelLocalPort++;
        if (this.nextTunnelLocalPort > this.portRangeMax) {
            this.nextTunnelLocalPort = this.portRangeMin;
        }

        session.setPortForwardingL("0.0.0.0", tunnelLocalPort, targetHost, targetPort);

        tunnel = new SSHTunnel(targetHost, targetPort, tunnelLocalPort, session);

        this.openTunnels.put(targetHost + targetPort, tunnel);

        SSHTunnelManagerImpl.logger.info("Created SSH tunnel to (" + targetHost + "," + targetPort + ") from local port "
            + tunnelLocalPort);

        return tunnelLocalPort;
    }

    public synchronized void destroySSHTunnel(final String targetHost, final int targetPort) throws Exception {
        SSHTunnel tunnel = this.openTunnels.get(targetHost + targetPort);
        if (tunnel != null) {
            tunnel.session.delPortForwardingL(tunnel.localPort);
            tunnel.session.disconnect();
            this.openTunnels.remove(targetHost + targetPort);
            SSHTunnelManagerImpl.logger.info("Destroyed SSH tunnel (" + targetHost + "," + targetPort + ")");
        }
    }

    public synchronized int getSSHTunnel(final String targetHost, final int targetPort) {
        SSHTunnel tunnel = this.openTunnels.get(targetHost + targetPort);
        if (tunnel != null) {
            return tunnel.localPort;
        }
        return -1;
    }

    synchronized void shutdown() {
        for (SSHTunnel tunnel : this.openTunnels.values()) {
            try {
                tunnel.session.delPortForwardingL(tunnel.localPort);
            } catch (JSchException e) {
            }
            tunnel.session.disconnect();
        }
    }
}
